#include <iostream>
#include <chrono>

#include "GL/glew.h"
#include "GL/glut.h"

#include "tools/textfile.h"
#include "tools/myBMP.h"

#define DATA_SIZE_X 512
#define DATA_SIZE_Y 512
#define WORK_GROUP_SIZE_X 16//number of threads per block in dimension X
#define WORK_GROUP_SIZE_Y 16//number of threads per block in dimension Y

using namespace std;
using namespace std::chrono;

void juliaCPU(int width, int height);

struct data
{
float x, y, z, w; 
};

void printShaderInfoLog(GLuint obj)
{
    int infologLength = 0;
    int charsWritten  = 0;
    char *infoLog;

	glGetShaderiv(obj, GL_INFO_LOG_LENGTH, &infologLength);

    if (infologLength > 0)
    {
        infoLog = (char *)malloc(infologLength);
        glGetShaderInfoLog(obj, infologLength, &charsWritten, infoLog);
		printf("%s\n",infoLog);
        free(infoLog);
    }
}

void printProgramInfoLog(GLuint obj)
{
    int infologLength = 0;
    int charsWritten  = 0;
    char *infoLog;

	glGetProgramiv(obj, GL_INFO_LOG_LENGTH, &infologLength);

    if (infologLength > 0)
    {
        infoLog = (char *)malloc(infologLength);
        glGetProgramInfoLog(obj, infologLength, &charsWritten, infoLog);
		printf("%s\n",infoLog);
        free(infoLog);
    }
}


void performComputation() 
{
	/*
	load program
	*/
	GLuint prog;

	char *cs = NULL;

	GLuint c = glCreateShader(GL_COMPUTE_SHADER);

	cs = textFileRead("shaders/shader.comp");

	const char * cc = cs;

	glShaderSource(c, 1, &cc,NULL);

	free(cs);
	
	glCompileShader(c);

	printShaderInfoLog(c);

	prog = glCreateProgram();
	glAttachShader(prog, c);

	glLinkProgram(prog);
	printProgramInfoLog(prog);

	/*
		Setting up the Shader Storage Buffer Objects
	*/

	GLuint dataSSbo;
	glGenBuffers( 1, &dataSSbo);
	glBindBuffer( GL_SHADER_STORAGE_BUFFER, dataSSbo );
	glBufferData( GL_SHADER_STORAGE_BUFFER, (DATA_SIZE_X*DATA_SIZE_Y) * sizeof(struct data), NULL, GL_STATIC_DRAW );

	GLint bufMask = GL_MAP_WRITE_BIT | GL_MAP_INVALIDATE_BUFFER_BIT ; // the invalidate makes a big difference when re-writing
	struct data *p = ( struct data *) glMapBufferRange(GL_SHADER_STORAGE_BUFFER, 0, (DATA_SIZE_X*DATA_SIZE_Y) * sizeof(struct data), bufMask);

	for( int i = 0; i < DATA_SIZE_X*DATA_SIZE_Y; i++ )
	{
		p[ i ].x = 0;
		p[ i ].y = 0;
		p[ i ].z = 0;
		p[ i ].w = 0;
	}

	glUnmapBuffer( GL_SHADER_STORAGE_BUFFER );

	glBindBufferBase( GL_SHADER_STORAGE_BUFFER, 4, dataSSbo );


	glUseProgram( prog );

	high_resolution_clock::time_point t0_gpu = high_resolution_clock::now();

	//number of work groups items per dimension
	glDispatchCompute( (DATA_SIZE_X + WORK_GROUP_SIZE_X - 1)  / WORK_GROUP_SIZE_X, (DATA_SIZE_Y + WORK_GROUP_SIZE_X - 1) / WORK_GROUP_SIZE_Y, 1 );
	glMemoryBarrier( GL_SHADER_STORAGE_BARRIER_BIT );

	glBindBuffer( GL_SHADER_STORAGE_BUFFER, dataSSbo );

	struct data *out = (struct data *) glMapBuffer( GL_SHADER_STORAGE_BUFFER, GL_READ_ONLY );
	glUnmapBuffer( GL_SHADER_STORAGE_BUFFER );

	high_resolution_clock::time_point t1_gpu = high_resolution_clock::now();

    auto duration = duration_cast<microseconds>( t1_gpu - t0_gpu ).count();
	cout << "GPU duration\n";
	cout << duration;
	cout << "\n";

	Bmp bmp; 
	bmp.width = DATA_SIZE_X;
	bmp.height = DATA_SIZE_Y;
	bmp.img = createBMP(bmp);

	for (int i = 0; i < bmp.width; i++)
		for (int j = 0; j < bmp.height; j++)
			setRGB(bmp, i, j, out[i + j*DATA_SIZE_Y].x, out[i + j*DATA_SIZE_Y].y, out[i + j*DATA_SIZE_Y].z);

	saveBMP("resultGPU.bmp", bmp);

	freeBMP(bmp);

	juliaCPU(DATA_SIZE_X, DATA_SIZE_Y);
	
	
}

void juliaCPU(int width, int height){

	struct cuComplex {
		float r;
		float i;
		cuComplex( float a, float b ) : r(a), i(b) {}

		float magnitude2( void ) {
			return r * r + i * i;
		}

		cuComplex operator*(const cuComplex& a) {
			return cuComplex(r*a.r - i*a.i, i*a.r + r*a.i);
		}

		cuComplex operator+(const cuComplex& a) {
			return cuComplex(r+a.r, i+a.i);
		}
	};

	Bmp bmp; 
	bmp.width = width;
	bmp.height = height;
	bmp.img = createBMP(bmp);

	high_resolution_clock::time_point t0_cpu = high_resolution_clock::now();

	const float scale = 1.5;
	const int max_iteration = 200;

	for (int i = 0; i < bmp.width; i++)
		for (int j = 0; j < bmp.height; j++) {
		
			float jx = scale * (float)(width/2 - i)/(width/2);
			float jy = scale * (float)(height/2 - j)/(height/2);

			cuComplex c(-0.8, 0.156);
			//cuComplex c(-0.7269, 0.1889);
			//cuComplex c(-0.4,0.6);
			cuComplex a(jx, jy);

			int iteration  = 0;
			

			while (iteration < max_iteration && a.magnitude2() < 4) {
				
				a = a * a + c;

				iteration ++;
			}

			if (iteration < 20)
				setRGB(bmp, i, j,  0, 0, 0);
			else if (iteration < 40)
				setRGB(bmp, i, j, 0.173,0.008,0.22);
			else if (iteration < 60)
				setRGB(bmp, i, j, 0.271,0.063,0.329);
			else if (iteration < 100)
				setRGB(bmp, i, j, 0.49,0.286,0.549);
			else if (iteration < 140)
				setRGB(bmp, i, j, 0.616,0.447,0.663);
			else setRGB(bmp, i, j, 0.376,0.157,0.439);
		
		}

		high_resolution_clock::time_point t1_cpu = high_resolution_clock::now();
		auto duration = duration_cast<microseconds>( t1_cpu - t0_cpu ).count();
		cout << "CPU duration\n";
		cout << duration;
		cout << "\n";

		saveBMP("resultCPU.bmp", bmp);

		freeBMP(bmp);

}


int main(int argc, char **argv) {

	// init GLUT and GLEW
	glutInit(&argc, argv);
	glutCreateWindow("OpenGL context");

	glewInit();
	if (glewIsSupported("GL_VERSION_4_3"))
	{
		std::cout << "OpenGL 4.3 is supported! \n";
	}

	performComputation() ;

	system( "PAUSE" );

    return 1;
}